package com.css.simulation.resource.scheduler.consumer;


import api.common.pojo.constants.DictConstants;
import api.common.pojo.dto.ProjectMessageDTO;
import api.common.util.*;
import com.css.simulation.resource.scheduler.mapper.*;
import com.css.simulation.resource.scheduler.pojo.po.*;
import com.css.simulation.resource.scheduler.pojo.to.*;
import com.css.simulation.resource.scheduler.util.MinioUtil;
import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import io.minio.MinioClient;
import lombok.SneakyThrows;
import lombok.extern.slf4j.Slf4j;
import org.apache.http.client.config.RequestConfig;
import org.apache.http.impl.client.CloseableHttpClient;
import org.apache.kafka.clients.consumer.ConsumerRecord;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.kafka.annotation.KafkaListener;
import org.springframework.kafka.core.KafkaTemplate;
import org.springframework.stereotype.Component;

import java.io.InputStream;
import java.util.*;
import java.util.concurrent.atomic.AtomicInteger;

@Component
@Slf4j
public class AutoProjectConsumer {

    private static final String USER_ID = "simulation-resource-scheduler";

    @Autowired
    CloseableHttpClient closeableHttpClient;
    @Autowired
    RequestConfig requestConfig;
    @Autowired
    MinioClient minioClient;
    @Value("${minio.bucket-name}")
    String bucketName;
    @Autowired
    KafkaTemplate<String, String> kafkaTemplate;
    @Autowired
    StringRedisTemplate redisTemplate;
    @Autowired
    ManualProjectMapper manualProjectMapper;
    @Autowired
    AutoProjectMapper autoProjectMapper;
    @Autowired
    AutoSubProjectMapper autoSubProjectMapper;
    @Autowired
    TaskMapper taskMapper;
    @Autowired
    IndexMapper indexMapper;
    @Autowired
    IndexTemplateMapper indexTemplateMapper;
    @Autowired
    SceneMapper sceneMapper;
    @Autowired
    VehicleMapper vehicleMapper;
    @Autowired
    SensorCameraMapper sensorCameraMapper;
    @Autowired
    SensorOgtMapper sensorOgtMapper;
    @Autowired
    AlgorithmMapper algorithmMapper;
    //    @Autowired
//    ApiClient apiClient;
    @Value("${scheduler.auto-project.topic}")
    String autoProjectTopic;
    @Value("${scheduler.manual-project.result-path-minio}")
    String resultPathMinio;
    @Value("${scheduler.manual-project.job-template}")
    String jobTemplate;
    @Value("${scheduler.manual-project.job-yaml}")
    String jobYaml;
    @Value("${scheduler.linux-temp-path}")
    String linuxTempPath;

    @KafkaListener(groupId = "simulation-resource-scheduler", topics = "${scheduler.manual-project.topic}")
    @SneakyThrows
    public void parseAutoProject(ConsumerRecord<String, String> projectRecord) {
        log.info("------- AutoProjectConsumer 接收到自动项目的子项目开始执行的消息为：" + projectRecord);
        //1 读取 kafka 的 project 信息
        /*
            {
                "projectId": "sadfasdfs",	// 项目 id
                "algorithmId": "sadfasdfs",	// 算法 id
                "vehicleConfigId": "sadfasdfs",	// 车辆 id
                "scenePackageId": "sadfasdfs",	// 场景包 id
                "maxSimulationTime": 11111,	// 最大仿真时间
                "parallelism": 30		// 并行度
            }
         */
        String projectJson = projectRecord.value();
        ProjectMessageDTO projectMessageDTO = JsonUtil.jsonToBean(projectJson, ProjectMessageDTO.class);
        String subProjectId = projectMessageDTO.getProjectId();    // 自动运行项目的子项目 id

        redisTemplate.opsForValue().set(autoProjectTopic + ":" + subProjectId + ":taskCompleted", "0");  // 设置项目已完成任务为 0
        autoProjectMapper.updateProjectStateBySubProjectId(subProjectId, DictConstants.PROJECT_RUNNING);
        autoSubProjectMapper.updateProjectStateById(subProjectId, DictConstants.PROJECT_RUNNING);   // 修改该 project 的状态为执行中，同时将已完成任务重置为 0 方便测试。


        // -------------------------------- 1 场景 --------------------------------
        // 根据 packageId，拿到场景 id，然后获取每个场景的文件地址。
        String packageId = projectMessageDTO.getScenePackageId();
        List<IndexTemplatePO> leafIndexList = indexTemplateMapper.selectLeafIndexByPackageId(packageId);
        log.info("------- AutoProjectConsumer 自动项目的子项目 " + subProjectId + " 的叶子指标为：" + leafIndexList);
        List<ScenePO> sceneList = new ArrayList<>();
        leafIndexList.forEach(leafIndex -> {
            List<String> naturalIdList = new ArrayList<>();
            List<String> standardIdList = new ArrayList<>();
            List<String> accidentIdList = new ArrayList<>();
            String naturalIds = leafIndex.getSceneNaturalIds();
            String standardIds = leafIndex.getSceneStatueIds();
            String accidentIds = leafIndex.getSceneTrafficIds();
            if (StringUtil.isNotEmpty(naturalIds)) {
                String[] naturalIdArray = naturalIds.split(",");
                naturalIdList.addAll(Arrays.asList(naturalIdArray));
            }
            if (StringUtil.isNotEmpty(standardIds)) {
                String[] standardArray = standardIds.split(",");
                standardIdList.addAll(Arrays.asList(standardArray));
            }
            if (StringUtil.isNotEmpty(accidentIds)) {
                String[] accidentIdArray = accidentIds.split(",");
                accidentIdList.addAll(Arrays.asList(accidentIdArray));
            }
            if (CollectionUtil.isNotEmpty(naturalIdList)) {
                sceneList.addAll(sceneMapper.selectNaturalByIdList(naturalIdList));
            }
            if (CollectionUtil.isNotEmpty(standardIdList)) {
                sceneList.addAll(sceneMapper.selectStandardByIdList(standardIdList));
            }
            if (CollectionUtil.isNotEmpty(accidentIdList)) {
                sceneList.addAll(sceneMapper.selectAccidentByIdList(accidentIdList));
            }
        });

        int taskNumber = sceneList.size();
        Set<ScenePO> sceneSet = new HashSet<>(sceneList);
        log.info("------- AutoProjectConsumer 共有 " + taskNumber + " 个任务，对应 " + sceneSet.size() + " 个场景！");
        redisTemplate.opsForValue().set(autoProjectTopic + ":" + subProjectId + ":taskNumber", taskNumber + "");
        // -------------------------------- 2 模型 --------------------------------
        // 根据车辆配置id vehicleConfigId, 获取 模型信息和传感器信息
        String vehicleConfigId = projectMessageDTO.getVehicleConfigId();// 模型配置 id
        VehiclePO vehiclePO = vehicleMapper.selectByVehicleConfigId(vehicleConfigId);   // 模型
        List<CameraPO> cameraPOList = sensorCameraMapper.selectCameraByVehicleConfigId(vehicleConfigId);    // 摄像头
        List<OgtPO> ogtPOList = sensorOgtMapper.selectOgtByVehicleId(vehicleConfigId); // 完美传感器

        // -------------------------------- 3 任务消息 --------------------------------
        AtomicInteger messageNumber = new AtomicInteger();
        // 根据场景创建任务，组装 task 消息
        int maxSimulationTime = projectMessageDTO.getMaxSimulationTime();

        for (ScenePO scenePO : sceneSet) {
            String sceneId = scenePO.getId();
            // 根据 projectId 和 sceneId 查询任务信息
            List<String> taskIdList = taskMapper.selectIdByProjectIdAndSceneId(subProjectId, sceneId);
            // 可能会存在多个指标下有同样的场景，所以会查出多个指标
            List<String> lastTargetIdList = indexMapper.selectLeafIndexIdByProjectAndSceneId(subProjectId, "%" + sceneId + "%");

            if (CollectionUtil.isEmpty(taskIdList)) {    // 如果没有旧任务说明是创建任务
                lastTargetIdList.forEach(lastTargetId -> {
                    String taskId = StringUtil.getRandomUUID();
                    String resultPath = resultPathMinio + subProjectId + "/" + taskId;
                    // 保存任务信息
                    TaskPO taskPO = TaskPO.builder() // run_start_time 和 run_end_time 不填
                            .id(taskId)
                            .pId(subProjectId)
                            .sceneId(sceneId)
                            .lastTargetId(lastTargetId)
                            .sceneName(scenePO.getName())
                            .sceneType(scenePO.getType())
                            .runState(DictConstants.TASK_PENDING)
                            .runResultFilePath(resultPath)
                            .build();
                    taskPO.setCreateTime(TimeUtil.getNowForMysql());
                    taskPO.setCreateUserId(USER_ID);
                    taskPO.setModifyTime(TimeUtil.getNowForMysql());
                    taskPO.setModifyUserId(USER_ID);
                    taskPO.setModifyTime(TimeUtil.getNowForMysql());
                    taskPO.setIsDeleted("0");
                    taskMapper.insert(taskPO);
                    // 心跳信息存在緩存中
                    redisTemplate.opsForValue().set(autoProjectTopic + ":" + subProjectId + ":" + taskId + ":tick", TimeUtil.getNowString());
                    // 组装 task 消息
                    TaskTO taskTO = TaskTO.builder()
                            .info(InfoTO.builder()
                                    .project_id(subProjectId)
                                    .task_id(taskId)
                                    .task_path(resultPath)
                                    .default_time(maxSimulationTime)
                                    .build())
                            .scenario(ScenarioTO.builder()
                                    .scenario_osc(scenePO.getScenarioOsc())
                                    .scenario_odr(scenePO.getScenarioOdr())
                                    .scenario_osgb(scenePO.getScenarioOsgb())
                                    .build())
                            .vehicle(VehicleTO.builder()
                                    .model(ModelTO.builder()
                                            .model_label(vehiclePO.getModelLabel())
                                            .build())
                                    .dynamics(DynamicsTO.builder()
                                            .dynamics_maxspeed(vehiclePO.getMaxSpeed())
                                            .dynamics_enginepower(vehiclePO.getEnginePower())
                                            .dynamics_maxdecel(vehiclePO.getMaxDeceleration())
                                            .dynamics_maxsteering(vehiclePO.getMaxSteeringAngle())
                                            .dynamics_mass(vehiclePO.getMass())
                                            .dynamics_frontsurfaceeffective(vehiclePO.getFrontSurfaceEffective())
                                            .dynamics_airdragcoefficient(vehiclePO.getAirDragCoefficient())
                                            .dynamics_rollingresistance(vehiclePO.getRollingResistanceCoefficient())
                                            .dynamics_wheeldiameter(vehiclePO.getWheelDiameter())
                                            .dynamics_wheeldrive(vehiclePO.getWheelDrive())
                                            .dynamics_overallefficiency(vehiclePO.getOverallEfficiency())
                                            .dynamics_distfront(vehiclePO.getFrontDistance())
                                            .dynamics_distrear(vehiclePO.getRearDistance())
                                            .dynamics_distleft(vehiclePO.getLeftDistance())
                                            .dynamics_distright(vehiclePO.getRightDistance())
                                            .dynamics_distheight(vehiclePO.getHeightDistance())
                                            .dynamics_wheelbase(vehiclePO.getWheelbase())
                                            .build())
                                    .sensors(SensorsTO.builder()   // 根据 vehicleId 查询绑定的传感器列表
                                            .camera(cameraPOList)
                                            .OGT(ogtPOList)
                                            .build())
                                    .build())
                            .build();

                    //4-4 将对象转成 json
                    String taskJson = "";
                    try {
                        taskJson = JsonUtil.beanToJson(taskTO);
                    } catch (JsonProcessingException e) {
                        e.printStackTrace();
                    }
                    //4-5 将 projectId 作为 topic 名称，发送 task 信息到 kafka
                    String finalTaskJson = taskJson;
                    kafkaTemplate.send(subProjectId, taskJson).addCallback(success -> {
                        messageNumber.addAndGet(1);
                        // 消息发送到的topic
                        assert success != null;
                        String topic = success.getRecordMetadata().topic();
                        // 消息发送到的分区
                        int partition = success.getRecordMetadata().partition();
                        // 消息在分区内的offset
                        long offset = success.getRecordMetadata().offset();
                        log.info("------- ManualProjectConsumer 发送消息成功：\n"
                                + "主题 topic 为：" + topic + "\n"
                                + "分区 partition 为：" + partition + "\n"
                                + "偏移量为：" + offset + "\n"
                                + "消息体为：" + finalTaskJson);
                    }, failure -> {
                        log.error("------- 发送消息失败：" + failure.getMessage());
                    });

                    redisTemplate.opsForValue().set(autoProjectTopic + ":" + subProjectId + ":" + taskId + ":message", taskJson);
                });

            } else {// 重新执行任务
                taskIdList.forEach(taskId -> {
                    String resultPath = resultPathMinio + subProjectId + "/" + taskId;
                    String lastTargetId = taskMapper.selectLastTargetIdById(taskId);
                    taskMapper.updateStateById(DictConstants.TASK_PENDING, taskId);
                    // 心跳信息存在緩存中
                    redisTemplate.opsForValue().set(autoProjectTopic + ":" + subProjectId + ":" + taskId + ":tick", TimeUtil.getNowString());
                    // 组装 task 消息
                    TaskTO taskTO = TaskTO.builder()
                            .info(InfoTO.builder()
                                    .project_id(subProjectId)
                                    .task_id(taskId)
                                    .task_path(resultPath)
                                    .default_time(maxSimulationTime)
                                    .build())
                            .scenario(ScenarioTO.builder()
                                    .scenario_osc(scenePO.getScenarioOsc())
                                    .scenario_odr(scenePO.getScenarioOdr())
                                    .scenario_osgb(scenePO.getScenarioOsgb())
                                    .build())
                            .vehicle(VehicleTO.builder()
                                    .model(ModelTO.builder()
                                            .model_label(vehiclePO.getModelLabel())
                                            .build())
                                    .dynamics(DynamicsTO.builder()
                                            .dynamics_maxspeed(vehiclePO.getMaxSpeed())
                                            .dynamics_enginepower(vehiclePO.getEnginePower())
                                            .dynamics_maxdecel(vehiclePO.getMaxDeceleration())
                                            .dynamics_maxsteering(vehiclePO.getMaxSteeringAngle())
                                            .dynamics_mass(vehiclePO.getMass())
                                            .dynamics_frontsurfaceeffective(vehiclePO.getFrontSurfaceEffective())
                                            .dynamics_airdragcoefficient(vehiclePO.getAirDragCoefficient())
                                            .dynamics_rollingresistance(vehiclePO.getRollingResistanceCoefficient())
                                            .dynamics_wheeldiameter(vehiclePO.getWheelDiameter())
                                            .dynamics_wheeldrive(vehiclePO.getWheelDrive())
                                            .dynamics_overallefficiency(vehiclePO.getOverallEfficiency())
                                            .dynamics_distfront(vehiclePO.getFrontDistance())
                                            .dynamics_distrear(vehiclePO.getRearDistance())
                                            .dynamics_distleft(vehiclePO.getLeftDistance())
                                            .dynamics_distright(vehiclePO.getRightDistance())
                                            .dynamics_distheight(vehiclePO.getHeightDistance())
                                            .dynamics_wheelbase(vehiclePO.getWheelbase())
                                            .build())
                                    .sensors(SensorsTO.builder()   // 根据 vehicleId 查询绑定的传感器列表
                                            .camera(cameraPOList)
                                            .OGT(ogtPOList)
                                            .build())
                                    .build())
                            .build();

                    //4-4 将对象转成 json
                    String taskJson = "";
                    try {
                        taskJson = JsonUtil.beanToJson(taskTO);
                    } catch (JsonProcessingException e) {
                        e.printStackTrace();
                    }
                    //4-5 将 projectId 作为 topic 名称，发送 task 信息到 kafka
                    String finalTaskJson = taskJson;
                    kafkaTemplate.send(subProjectId, taskJson).addCallback(success -> {
                        messageNumber.addAndGet(1);
                        // 消息发送到的topic
                        assert success != null;
                        String topic = success.getRecordMetadata().topic();
                        // 消息发送到的分区
                        int partition = success.getRecordMetadata().partition();
                        // 消息在分区内的offset
                        long offset = success.getRecordMetadata().offset();
                        log.info("------- ManualProjectConsumer 发送消息成功：\n"
                                + "主题 topic 为：" + topic + "\n"
                                + "分区 partition 为：" + partition + "\n"
                                + "偏移量为：" + offset + "\n"
                                + "消息体为：" + finalTaskJson);
                    }, failure -> {
                        log.error("------- ManualProjectConsumer 发送消息失败：" + failure.getMessage());
                    });
                    redisTemplate.opsForValue().set(autoProjectTopic + ":" + subProjectId + ":" + taskId + ":message", taskJson);
                });
            }
        }

        log.info("------- ManualProjectConsumer 共发送了" + messageNumber.get() + " 条消息！");

        // -------------------------------- 4 算法导入（一期按单机版做） --------------------------------
        // 私有仓库导入算法镜像
        String algorithmId = projectMessageDTO.getAlgorithmId();    // 算法 id
        //4-1 根据算法 id 获取算法文件地址、是否已导入成镜像。
        AlgorithmPO algorithmPO = algorithmMapper.selectById(algorithmId);
        String algorithmTarLinuxTempPath = linuxTempPath + "algorithm/" + algorithmId + ".tar";
        String dockerImage = "algorithm_" + algorithmId + ":latest";
        if (algorithmPO == null) {
            // 访问索为远程接口
            String tokenUri = "http://open-api.zoogooy.com/cgi-bin/token/token?grant_type=client_credential";
            String appid = "3e64be4a29e5478f9717d53c11ab26ad";
            String secret = "f183079f97ac9ed81a864619a83fc17a";
            String tokenUrl = tokenUri + "&appid=" + appid + "&secret=" + secret;
            String result = HttpUtil.get(closeableHttpClient, requestConfig, tokenUrl);
            ObjectMapper objectMapper = new ObjectMapper();
            JsonNode jsonNode = objectMapper.readTree(result);
            JsonNode dataNode = jsonNode.path("data");

            String token = dataNode.path("access_token").asText();

            /*
            {
              "data" : {
                "totalElements" : 19,
                "totalPage" : null,
                "currentPage" : 0,
                "size" : 0,
                "content" : [ {
                  "id" : "ICV20220418170905",
                  "algorithmId" : "2022041813435600064848bbc0e0efb441978d57",
                  "algorithmName" : "SUAN FA2",
                  "description" : "<div class=\"index-module_textWrap_3ygOc \">\n<p>4月17日，中科院图书馆处某位负责落实此事的老师向红星新闻证实，网传的近“千万级别”续订费用和停用知网访问一事属实，中科院各大所正积极应对因停订知网带来的影响。</p>\n<p>&nbsp;</p>\n</div>\n<div class=\"index-module_textWrap_3ygOc \">\n<p>&nbsp;</p>\n<p>“自2022年4月20日起，停用CNKI数据库。以后会用万方和维普替代，查不到的文献可以用文献传递等方式。”至于何时会恢复对知网的访问，该老师表示，目前并不确定，“至少今年应该没有这个计划了”。</p>\n<p>&nbsp;</p>\n</div>\n<div class=\"index-module_textWrap_3ygOc \">\n<p>&nbsp;</p>\n<p>去年底，89岁退休教授赵德馨起诉中国知网，引发热议。原因是后者擅自收录他的100多篇论文，赵德馨没拿到一分钱稿费，自己下载还要付费。</p>\n<p>&nbsp;</p>\n</div>\n<div class=\"index-module_textWrap_3ygOc \">\n<p>&nbsp;</p>\n<p>最终赵德馨全部胜诉，累计获赔70余万元。中国知网不再收录他的文章，已收录的也全部下架。</p>\n<p>&nbsp;</p>\n</div>\n<div class=\"index-module_textWrap_3ygOc \">\n<p>&nbsp;</p>\n<p>据中新经纬2021年报道，博士论文、硕士论文在中国知网出版，作者本人最高仅可获得100元现金，以及400元面值的检索阅读卡作为稿酬。但是作者的论文每在中国知网上被下载一次，平台就会收取15元/本甚至25元/本的费用。</p>\n<p>&nbsp;</p>\n</div>\n<div class=\"index-module_textWrap_3ygOc \">\n<p>&nbsp;</p>\n<p>在这种模式下，中国知网吸金不少。财报数据显示，2020年中国知网年收入近12亿元，毛利率近54%。</p>\n<p>&nbsp;</p>\n</div>\n<div class=\"index-module_textWrap_3ygOc \">\n<p>&nbsp;</p>\n<p>天眼查信息显示，中国知网所属公司为同方知网(北京)技术有限公司(下称同方知网)，该公司成立于2004年11月18日，是A股上市公司同方股份的子公司。</p>\n<p>&nbsp;</p>\n</div>\n<div class=\"index-module_textWrap_3ygOc \">\n<p>&nbsp;</p>\n<p>目前，中国知网已经建成了世界上全文信息量规模最大的“CNKI数字图书馆”，成为中国高校师生最常用的文献资料数据库。</p>\n<p>&nbsp;</p>\n</div>\n<div class=\"index-module_textWrap_3ygOc \">\n<p>&nbsp;</p>\n<p>据同方股份年报，同方知网主要从事互联网出版与服务业务，目前已经形成“中国知网”(CNKI)门户网站，为用户提供《中国知识资源总库》《中国学术期刊数据库》《中国博硕士论文数据库》、《中国年鉴全文数据库》《中国工具书网络出版总库》等一系列产品，以及知识资源互联网搜索、共享和网络出版服务。同方股份财报显示，2020年全年，同方知网主营业务收入11.68亿元、归母净利润1.93亿元，毛利率53.93%；2021年上半年，该公司主营业务收入4.96亿元、归母净利润1892.70万元，毛利率为51.30%。据悉，中国知网曾因频繁涨价遭到高校抵制。2016年1月，武汉理工大学发布了知网停用的通知。校方称：“由于续订价格涨价离谱，我校与中国知网公司的谈判不成功。这些年来，CNKI公司涨价幅度过大的行为已经收到全国很多高校的抵制，包括许多知名的985高校。”武汉理工大学图书馆称，2000年以来，知网每年的报价涨幅都超过10%，从2010年到2016年的报价涨幅为132.86%，年平均涨幅为18.98%。但在不到1个月后，武汉理工大学又重新订购并恢复开通中国知网数据。</p>\n</div>",
                  "algorithmType" : "2"
                }, {
                  "id" : "ICV20220408165152",
                  "algorithmId" : "202204081650140005916540f84c37764591a5c4",
                  "algorithmName" : "ceshi 33",
                  "description" : "",
                  "algorithmType" : "2"
                } ],
                "hasPreviousPage" : false,
                "hasNextPage" : false
              },
              "success" : true,
              "message" : "ok",
              "code" : 1,
              "nowTime" : "2022-04-22 10:14:40"
            }
             */
            String downloadUrl = "http://open-api.zoogooy.com/cgi-bin/api/icv-algorithm-agg/simulation/download"
                    + "?access_token=" + token
                    + "&id=" + algorithmId;
            String tempDownloadUrl = HttpUtil.get(closeableHttpClient, requestConfig, downloadUrl);
            InputStream inputStream = HttpUtil.getInputStream(closeableHttpClient, requestConfig, tempDownloadUrl);
            FileUtil.writeInputStreamToLocalFile(inputStream, algorithmTarLinuxTempPath);
            LinuxUtil.execute("docker import " + algorithmTarLinuxTempPath + " " + dockerImage);
        } else {
            String minioPath = algorithmPO.getMinioPath();
            if ("0".equals(algorithmPO.getDockerImport()) || algorithmPO.getDockerImport() == null) {
                // 下载算法文件到本地（ 2 到仓库服务器）
                MinioUtil.downloadToFile(minioClient, bucketName, minioPath, algorithmTarLinuxTempPath);
                //4-2 本地执行 docker load 算法文件成镜像（集群版可改成用 docker-java 操作仓库）
                LinuxUtil.execute("docker import " + algorithmTarLinuxTempPath + " " + dockerImage);
                algorithmMapper.updateDockerImportAndDockerImageById("1", dockerImage, algorithmId);
            } else if ("1".equals(algorithmPO.getDockerImport()) && StringUtil.isNotEmpty(algorithmPO.getDockerImage())) {
                dockerImage = algorithmPO.getDockerImage();
            } else {
                throw new RuntimeException("算法 " + algorithmId + " 的 mysql 数据有误！");
            }
        }

        // -------------------------------- 5 创建 pod 开始执行 --------------------------------
        int completions = sceneList.size();     // 结束标
        int parallelism = projectMessageDTO.getParallelism();    // 并行度
        log.info("------- AutoProjectConsumer 项目 " + subProjectId + " 的完成度为：" + completions);
        log.info("------- AutoProjectConsumer 项目 " + subProjectId + " 的并行度为：" + parallelism);
        String jobTemplateYamlPathSource = jobTemplate + "job-template.yaml";
        String jobTemplateYamlPathTarget = jobYaml + "project-" + subProjectId + ".yaml";
        String yamlSource = FileUtil.read(jobTemplateYamlPathSource);
        log.info("------- ManualProjectConsumer 模板文件为：" + yamlSource);
        String replace0 = yamlSource.replace("job-cloud-simulation", "project-" + subProjectId);
        String replace1 = replace0.replace("vtd-container", "vtd-" + subProjectId);
        String replace2 = replace1.replace("algorithm-container", "algorithm-" + subProjectId);
        String replace3 = replace2.replace("algorithm-image", dockerImage);
        String replace4 = replace3.replace("projectId", subProjectId);
        String replace5 = replace4.replace("completions-number", completions + "");
        String replace6 = replace5.replace("parallelism-number", parallelism + "");
        String replace7 = replace6.replace("apiVers1on", "apiVersion");
        String replace8 = replace7.replace("1atch/v1", "batch/v1");
        log.info("------- ManualProjectConsumer 开始执行 yaml 文件" + replace8);
        FileUtil.writeStringToLocalFile(replace8, jobTemplateYamlPathTarget);
        LinuxUtil.execute("kubectl apply -f " + jobTemplateYamlPathTarget);
    }

    @KafkaListener(groupId = "simulation-resource-scheduler", topics = "${scheduler.manual-project.stop-topic}")
    @SneakyThrows
    public void stopAutoProject(ConsumerRecord<String, String> stopRecord) {
        log.info("------- ManualProjectConsumer 接收到的项目终止消息为：" + stopRecord);
        //1 读取 kafka 的项目停止信息
        /*
            {
                "projectId": "sadfasdfs",	// 项目 id
            }
         */
        String json = stopRecord.value();
        ObjectMapper objectMapper = new ObjectMapper();
        JsonNode jsonNode = objectMapper.readTree(json);
        String projectId = jsonNode.path("projectId").asText();
        manualProjectMapper.updateProjectState(projectId, DictConstants.PROJECT_TERMINATED, TimeUtil.getNowForMysql());
        LinuxUtil.execute("kubectl delete job project-" + projectId);
        redisTemplate.delete(autoProjectTopic + ":" + projectId + ":check");
    }



}
