package com.css.simulation.resource.scheduler.scheduler;

import api.common.pojo.constants.DictConstants;
import api.common.util.SshUtil;
import api.common.util.StringUtil;
import api.common.util.TimeUtil;
import com.css.simulation.resource.scheduler.mapper.ManualProjectMapper;
import com.css.simulation.resource.scheduler.mapper.TaskMapper;
import com.css.simulation.resource.scheduler.pojo.po.ProjectPO;
import com.css.simulation.resource.scheduler.pojo.po.TaskPO;
import com.css.simulation.resource.scheduler.service.TaskService;
import io.kubernetes.client.openapi.ApiClient;
import lombok.extern.slf4j.Slf4j;
import org.apache.sshd.client.SshClient;
import org.apache.sshd.client.session.ClientSession;
import org.springframework.beans.factory.annotation.Autowired;
import org.springframework.beans.factory.annotation.Value;
import org.springframework.data.redis.core.StringRedisTemplate;
import org.springframework.kafka.core.KafkaTemplate;
import org.springframework.scheduling.annotation.Scheduled;
import org.springframework.stereotype.Component;

import java.io.IOException;
import java.util.List;
import java.util.Objects;

@Component
@Slf4j
public class ProjectScheduler {

    @Value("${scheduler.manual-project.topic}")
    String manualProjectTopic;
    @Autowired
    StringRedisTemplate redisTemplate;
    @Autowired
    TaskService taskService;
    @Autowired
    TaskMapper taskMapper;
    @Autowired
    ManualProjectMapper manualProjectMapper;
    @Value("${scheduler.manual-project.job-yaml}")
    String jobYaml;
    @Value("${scheduler.score.hostname}")
    String hostname;
    @Value("${scheduler.score.username}")
    String username;
    @Value("${scheduler.score.password}")
    String password;
    @Autowired
    ApiClient apiClient;
    @Autowired
    KafkaTemplate<String, String> kafkaTemplate;


//    /**
//     * 调度项目启动
//     * @throws IOException 超时时间
//     */
//    @Scheduled(fixedDelay = 60 * 1000)
//    public void dispatchProject() throws IOException {
//        //1 等待执行的项目
//
//
//
//        long timeout = 2 * 60 * 1000L;
//
//        SshClient client = SshUtil.getClient();
//        ClientSession session = SshUtil.getSession(client, hostname, username, password);
//        List<TaskPO> executingTaskList = taskMapper.selectExecuting();
//
//        log.info("ProjectScheduler--taskTimeout 正在运行的任务有：" + executingTaskList);
//        if (executingTaskList != null && executingTaskList.size() > 0) {
//            for (TaskPO task : executingTaskList) {
//                String taskId = task.getId();
//                String projectId = task.getPId();
//                long lastTickTime = Long.parseLong(Objects.requireNonNull(redisTemplate.opsForValue().get(manualProjectTopic + ":" + projectId + ":" + taskId + ":tick")));
//                if (TimeUtil.getNow() - lastTickTime > timeout) {
//                    String podName = redisTemplate.opsForValue().get(manualProjectTopic + ":" + projectId + ":" + taskId + ":pod");
//                    taskService.taskState(taskId, DictConstants.TASK_ABORTED, podName);
//                }
//            }
//        }
//        session.close();
//        client.stop();
//    }


    /**
     * 处理 pod 超时
     * 同时也可处理 pod 莫名关闭，因为关闭之后也会超时
     *
     * @throws IOException 超时时间
     */
    @Scheduled(fixedDelay = 60 * 1000)
    public void taskTimeout() throws IOException {

        long timeout = 2 * 60 * 1000L;

        SshClient client = SshUtil.getClient();
        ClientSession session = SshUtil.getSession(client, hostname, username, password);
        List<TaskPO> executingTaskList = taskMapper.selectByRunState(DictConstants.TASK_RUNNING);
        log.info("ProjectScheduler--taskTimeout 正在运行的任务有：" + executingTaskList);
        if (executingTaskList != null && executingTaskList.size() > 0) {
            for (TaskPO task : executingTaskList) {
                String taskId = task.getId();
                String projectId = task.getPId();
                String userId = task.getCreateUserId();
                long lastTickTime = Long.parseLong(Objects.requireNonNull(redisTemplate.opsForValue().get(manualProjectTopic+ ":" + userId + ":" + projectId + ":task" + taskId + ":tick")));
                if (TimeUtil.getNow() - lastTickTime > timeout) {
                    String podName = redisTemplate.opsForValue().get(manualProjectTopic+ ":" + userId + ":" + projectId + ":task:" + taskId + ":pod");
                    taskService.taskState(taskId, DictConstants.TASK_ABORTED, podName);
                }
            }
        }
        session.close();
        client.stop();
    }

    /**
     * 解决 pod 莫名全部关闭但是 job 还在的问题
     * 检查如果有 job 在运行但是 pod 全部关闭的情况，此时需要重启一下 job
     */
    @Scheduled(fixedDelay = 30 * 1000)
    public void projectCheck() throws IOException {
        SshClient client = SshUtil.getClient();
        ClientSession session = SshUtil.getSession(client, hostname, username, password);

        //1 查询出正在运行中的 project
        List<ProjectPO> projectIdList = manualProjectMapper.selectByNowRunState(DictConstants.PROJECT_RUNNING);
        log.info("ProjectScheduler--projectCheck 运行中的项目有：" + projectIdList);
        //2 根据 projectId 获取 pod
        projectIdList.forEach(project -> {
            String projectId = project.getId();
            String userId = project.getCreateUserId();
            try {
                String checkKey = manualProjectTopic+ ":" + userId + ":" + projectId + ":check";
                String lastNowString = redisTemplate.opsForValue().get(checkKey);
                String podList = SshUtil.execute(session, "kubectl get pod | grep project-" + projectId);
                log.info("ProjectScheduler--projectCheck 项目 " + projectId + " 正在运行的 pod 为：\n" + podList);
                int taskNumber = StringUtil.countSubString(podList, "project");
                if (StringUtil.isEmpty(lastNowString) && taskNumber == 0) {
                    redisTemplate.opsForValue().set(checkKey, TimeUtil.getNowString());
                }
                if (StringUtil.isNotEmpty(lastNowString) && taskNumber == 0) {
                    // 判断两次是否超过2分钟
                    //3 如果 pod 为空，则重启 job
                    long lastNow = Long.parseLong(lastNowString);
                    long now = Long.parseLong(TimeUtil.getNowString());

                    if (now - lastNow > (long) 120 * 1000) {
                        redisTemplate.opsForValue().set(checkKey, TimeUtil.getNowString());
                        SshUtil.execute(session, "kubectl delete job project-" + projectId);
                        Thread.sleep(15000);
                        while (true) {
                            log.info("ProjectScheduler--projectCheck 准备重启项目 " + projectId);
                            String podList2 = SshUtil.execute(session, "kubectl get pod | grep project-" + projectId);
                            log.info("ProjectScheduler--projectCheck 项目 " + projectId + " 剩余的 pod 信息为：\n" + podList2);
                            int taskNumber2 = StringUtil.countSubString(podList2, "project");
                            if (taskNumber2 == 0) {
                                break;
                            }
                        }
                        Thread.sleep(15000);
                        log.info("ProjectScheduler--projectCheck 重新执行项目" + projectId);
                        String jobTemplateYamlPathTarget = jobYaml + "project-" + projectId + ".yaml";
                        SshUtil.execute(session, "kubectl apply -f " + jobTemplateYamlPathTarget);
                    }
                }
            } catch (IOException | InterruptedException e) {
                e.printStackTrace();
            }
        });

        session.close();
        client.stop();

    }
}
